/*
 * bsc1204167_drivers_misc_sgi_gru_grufault
 *
 * Fix for CVE-2022-3424, bsc#1204167
 *
 *  Copyright (c) 2023 SUSE
 *  Author: Marcos Paulo de Souza <mpdesouza@suse.com>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_SGI_GRU)

#if !IS_MODULE(CONFIG_SGI_GRU)
#error "Live patch supports only CONFIG=m"
#endif

/* klp-ccp: from drivers/misc/sgi-gru/grufault.c */
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/spinlock.h>
#include <linux/mm.h>
#include <linux/hugetlb.h>
#include <linux/io.h>
#include <linux/uaccess.h>
#include <asm/pgtable.h>

#include "gru.h"

/* klp-ccp: from drivers/misc/sgi-gru/gru.h */
#define GRU_HANDLE_STRIDE		256
#define GRU_CB_BASE			0

/* klp-ccp: from drivers/misc/sgi-gru/grutables.h */
#include <linux/mutex.h>
#include <linux/wait.h>
#include <linux/mmu_notifier.h>

/* klp-ccp: from drivers/misc/sgi-gru/grulib.h */
enum {sco_gseg_owner, sco_cch_req_slice, sco_blade_chiplet};
struct gru_set_context_option_req {
	unsigned long	gseg;
	int		op;
	int		val0;
	long		val1;
};

/* klp-ccp: from drivers/misc/sgi-gru/gru_instructions.h */
#if defined(CONFIG_IA64)
#error "klp-ccp: non-taken branch"
#elif defined(CONFIG_X86_64)
#include <asm/cacheflush.h>

#else
#error "klp-ccp: non-taken branch"
#endif

struct gru_instruction_bits;

/* klp-ccp: from drivers/misc/sgi-gru/gruhandles.h */
#define GRU_GSEG0_BASE		0
#define GRU_MCS_BASE		(64 * 1024 * 1024)

#define GRU_NUM_CB		128

#define GRU_TFH_BASE		(GRU_MCS_BASE + 0x18000)

#define GRU_CHIPLETS_PER_HUB	2
#define GRU_HUBS_PER_BLADE	1
#define GRU_CHIPLETS_PER_BLADE	(GRU_HUBS_PER_BLADE * GRU_CHIPLETS_PER_HUB)

static inline void *get_gseg_base_address(void *base, int ctxnum)
{
	return (void *)(base + GRU_GSEG0_BASE + GRU_GSEG_STRIDE * ctxnum);
}

static inline void *get_gseg_base_address_cb(void *base, int ctxnum, int line)
{
	return (void *)(get_gseg_base_address(base, ctxnum) +
			GRU_CB_BASE + GRU_HANDLE_STRIDE * line);
}

static inline struct gru_tlb_fault_handle *get_tfh(void *base, int ctxnum)
{
	return (struct gru_tlb_fault_handle *)(base + GRU_TFH_BASE +
					ctxnum * GRU_HANDLE_STRIDE);
}

static inline unsigned long get_cb_number(void *cb)
{
	return (((unsigned long)cb - GRU_CB_BASE) % GRU_GSEG_PAGESIZE) /
					GRU_HANDLE_STRIDE;
}

/* klp-ccp: from drivers/misc/sgi-gru/grutables.h */
static struct gru_stats_s (*klpe_gru_stats);
static struct gru_blade_state *(*klpe_gru_base)[];

#define GRU_MAX_BLADES		MAX_NUMNODES
#define GRU_MAX_GRUS		(GRU_MAX_BLADES * GRU_CHIPLETS_PER_BLADE)

struct gru_mm_tracker {				/* pack to reduce size */
	unsigned int		mt_asid_gen:24;	/* ASID wrap count */
	unsigned int		mt_asid:24;	/* current base ASID for gru */
	unsigned short		mt_ctxbitmap:16;/* bitmap of contexts using
						   asid */
} __attribute__ ((packed));

struct gru_mm_struct {
	struct mmu_notifier	ms_notifier;
	spinlock_t		ms_asid_lock;	/* protects ASID assignment */
	atomic_t		ms_range_active;/* num range_invals active */
	wait_queue_head_t	ms_wait_queue;
	DECLARE_BITMAP(ms_asidmap, GRU_MAX_GRUS);
	struct gru_mm_tracker	ms_asids[GRU_MAX_GRUS];
};

#define get_tfh_by_index(g, i)						\
	((struct gru_tlb_fault_handle *)get_tfh((g)->gs_gru_base_vaddr, (i)))

#define thread_cbr_number(gts, n) ((gts)->ts_cbr_idx[(n) / GRU_CBR_AU_SIZE] \
				  * GRU_CBR_AU_SIZE + (n) % GRU_CBR_AU_SIZE)

static struct gru_thread_state *(*klpe_gru_alloc_thread_state)(struct vm_area_struct
				*vma, int tsid);

static void (*klpe_gru_unload_context)(struct gru_thread_state *gts, int savestate);
static int (*klpe_gru_update_cch)(struct gru_thread_state *gts);

static struct vm_area_struct *(*klpe_gru_find_vma)(unsigned long vaddr);

static unsigned long (*klpe_gru_options);

static struct gru_thread_state *(*klpe_gru_find_lock_gts)(unsigned long vaddr);

static struct gru_thread_state *klpr_gru_alloc_locked_gts(unsigned long vaddr)
{
	struct mm_struct *mm = current->mm;
	struct vm_area_struct *vma;
	struct gru_thread_state *gts = ERR_PTR(-EINVAL);

	down_write(&mm->mmap_sem);
	vma = (*klpe_gru_find_vma)(vaddr);
	if (!vma)
		goto err;

	gts = (*klpe_gru_alloc_thread_state)(vma, TSID(vaddr, vma));
	if (IS_ERR(gts))
		goto err;
	mutex_lock(&gts->ts_ctxlock);
	downgrade_write(&mm->mmap_sem);
	return gts;

err:
	up_write(&mm->mmap_sem);
	return gts;
}

static void gru_unlock_gts(struct gru_thread_state *gts)
{
	mutex_unlock(&gts->ts_ctxlock);
	up_read(&current->mm->mmap_sem);
}

static int (*klpe_gru_try_dropin)(struct gru_state *gru,
			  struct gru_thread_state *gts,
			  struct gru_tlb_fault_handle *tfh,
			  struct gru_instruction_bits *cbk);

static int klpr_gru_user_dropin(struct gru_thread_state *gts,
			   struct gru_tlb_fault_handle *tfh,
			   void *cb)
{
	struct gru_mm_struct *gms = gts->ts_gms;
	int ret;

	gts->ustats.upm_tlbmiss++;
	while (1) {
		wait_event(gms->ms_wait_queue,
			   atomic_read(&gms->ms_range_active) == 0);
		prefetchw(tfh);	/* Helps on hdw, required for emulator */
		ret = (*klpe_gru_try_dropin)(gts->ts_gru, gts, tfh, cb);
		if (ret <= 0)
			return ret;
		KLPR_STAT(call_os_wait_queue);
	}
}

int klpp_gru_handle_user_call_os(unsigned long cb)
{
	struct gru_tlb_fault_handle *tfh;
	struct gru_thread_state *gts;
	void *cbk;
	int ucbnum, cbrnum, ret = -EINVAL;

	KLPR_STAT(call_os);

	/* sanity check the cb pointer */
	ucbnum = get_cb_number((void *)cb);
	if ((cb & (GRU_HANDLE_STRIDE - 1)) || ucbnum >= GRU_NUM_CB)
		return -EINVAL;

again:
	gts = (*klpe_gru_find_lock_gts)(cb);
	if (!gts)
		return -EINVAL;
	gru_dbg(grudev, "address 0x%lx, gid %d, gts 0x%p\n", cb, gts->ts_gru ? gts->ts_gru->gs_gid : -1, gts);

	if (ucbnum >= gts->ts_cbr_au_count * GRU_CBR_AU_SIZE)
		goto exit;

	if (klpp_gru_check_context_placement(gts)) {
		gru_unlock_gts(gts);
		(*klpe_gru_unload_context)(gts, 1);
		goto again;
	}

	/*
	 * CCH may contain stale data if ts_force_cch_reload is set.
	 */
	if (gts->ts_gru && gts->ts_force_cch_reload) {
		gts->ts_force_cch_reload = 0;
		(*klpe_gru_update_cch)(gts);
	}

	ret = -EAGAIN;
	cbrnum = thread_cbr_number(gts, ucbnum);
	if (gts->ts_gru) {
		tfh = get_tfh_by_index(gts->ts_gru, cbrnum);
		cbk = get_gseg_base_address_cb(gts->ts_gru->gs_gru_base_vaddr,
				gts->ts_ctxnum, ucbnum);
		ret = klpr_gru_user_dropin(gts, tfh, cbk);
	}
exit:
	gru_unlock_gts(gts);
	return ret;
}

int klpp_gru_set_context_option(unsigned long arg)
{
	struct gru_thread_state *gts;
	struct gru_set_context_option_req req;
	int ret = 0;

	KLPR_STAT(set_context_option);
	if (copy_from_user(&req, (void __user *)arg, sizeof(req)))
		return -EFAULT;
	gru_dbg(grudev, "op %d, gseg 0x%lx, value1 0x%lx\n", req.op, req.gseg, req.val1);

	gts = (*klpe_gru_find_lock_gts)(req.gseg);
	if (!gts) {
		gts = klpr_gru_alloc_locked_gts(req.gseg);
		if (IS_ERR(gts))
			return PTR_ERR(gts);
	}

	switch (req.op) {
	case sco_blade_chiplet:
		/* Select blade/chiplet for GRU context */
		if (req.val0 < -1 || req.val0 >= GRU_CHIPLETS_PER_HUB ||
		    req.val1 < -1 || req.val1 >= GRU_MAX_BLADES ||
		    (req.val1 >= 0 && !(*klpe_gru_base)[req.val1])) {
			ret = -EINVAL;
		} else {
			gts->ts_user_blade_id = req.val1;
			gts->ts_user_chiplet_id = req.val0;
			if (klpp_gru_check_context_placement(gts)) {
				gru_unlock_gts(gts);
				(*klpe_gru_unload_context)(gts, 1);
				return ret;
			}
		}
		break;
	case sco_gseg_owner:
 		/* Register the current task as the GSEG owner */
		gts->ts_tgid_owner = current->tgid;
		break;
	case sco_cch_req_slice:
 		/* Set the CCH slice option */
		gts->ts_cch_req_slice = req.val1 & 3;
		break;
	default:
		ret = -EINVAL;
	}
	gru_unlock_gts(gts);

	return ret;
}



#define LP_MODULE "gru"

#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1204167.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "gru_alloc_thread_state", (void *)&klpe_gru_alloc_thread_state,
	  "gru" },
	{ "gru_base", (void *)&klpe_gru_base, "gru" },
	{ "gru_find_lock_gts", (void *)&klpe_gru_find_lock_gts, "gru" },
	{ "gru_find_vma", (void *)&klpe_gru_find_vma, "gru" },
	{ "gru_options", (void *)&klpe_gru_options, "gru" },
	{ "gru_stats", (void *)&klpe_gru_stats, "gru" },
	{ "gru_try_dropin", (void *)&klpe_gru_try_dropin, "gru" },
	{ "gru_unload_context", (void *)&klpe_gru_unload_context, "gru" },
	{ "gru_update_cch", (void *)&klpe_gru_update_cch, "gru" },
};

static int bsc1204167_drivers_misc_sgi_gru_grufault_module_notify(struct notifier_block *nb,
					unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LP_MODULE))
		return 0;
	mutex_lock(&module_mutex);
	ret = __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
	mutex_unlock(&module_mutex);

	WARN(ret, "%s: delayed kallsyms lookup failed. System is broken and can crash.\n",
		__func__);

	return ret;
}

static struct notifier_block module_nb = {
	.notifier_call = bsc1204167_drivers_misc_sgi_gru_grufault_module_notify,
	.priority = INT_MIN+1,
};

int bsc1204167_drivers_misc_sgi_gru_grufault_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(LP_MODULE)) {
		ret = __klp_resolve_kallsyms_relocs(klp_funcs,
						    ARRAY_SIZE(klp_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void bsc1204167_drivers_misc_sgi_gru_grufault_cleanup(void)
{
	unregister_module_notifier(&module_nb);
}

#endif /* IS_ENABLED(CONFIG_SGI_GRU) */
